SystemOrganization addCategory: #Astar!

Object subclass: #Map
	instanceVariableNames: 'area pos_robot pos_goal solved solutionPath'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Astar'!

!Map commentStamp: 'MorganePigny 1/4/2017 10:02' prior: 0!
I'm a map represented by a grid which contains numbers associated with the environment like so: 
0 = GROUND
1 = WALL
2 = TRAP
3 = ROBOT DEPARTURE POSITION
4 = GOAL!

!Map class methodsFor: 'examples' stamp: 'MorganePigny 1/4/2017 16:32'!
map1
	<sampleInstance>
	| instance area  | 
	instance := self new.
	area := instance area. 
	area atRow: 1 atColumn: 1 put: 1.
	area atRow: 1 atColumn: 2 put: 0.
	area atRow: 1 atColumn: 3 put: 0.
	area atRow: 1 atColumn: 4 put: 0.
	area atRow: 1 atColumn: 5 put: 0.
	area atRow: 1 atColumn: 6 put: 0.
	area atRow: 1 atColumn: 7 put: 1.
	area atRow: 2 atColumn: 1 put: 0.
	area atRow: 2 atColumn: 2 put: 3.
	area atRow: 2 atColumn: 3 put: 1.
	area atRow: 2 atColumn: 4 put: 0.
	area atRow: 2 atColumn: 5 put: 0.
	area atRow: 2 atColumn: 6 put: 0.
	area atRow: 2 atColumn: 7 put: 1.
	area atRow: 3 atColumn: 1 put: 0.
	area atRow: 3 atColumn: 2 put: 1.
	area atRow: 3 atColumn: 3 put: 1.
	area atRow: 3 atColumn: 4 put: 0.
	area atRow: 3 atColumn: 5 put: 0.
	area atRow: 3 atColumn: 6 put: 0.
	area atRow: 3 atColumn: 7 put: 0.
	area atRow: 4 atColumn: 1 put: 0.
	area atRow: 4 atColumn: 2 put: 2.
	area atRow: 4 atColumn: 3 put: 0.
	area atRow: 4 atColumn: 4 put: 0.
	area atRow: 4 atColumn: 5 put: 0.
	area atRow: 4 atColumn: 6 put: 0.
	area atRow: 4 atColumn: 7 put: 0.
	area atRow: 5 atColumn: 1 put: 0.
	area atRow: 5 atColumn: 2 put: 0.
	area atRow: 5 atColumn: 3 put: 0.
	area atRow: 5 atColumn: 4 put: 0.
	area atRow: 5 atColumn: 5 put: 0.
	area atRow: 5 atColumn: 6 put: 0.
	area atRow: 5 atColumn: 7 put: 0.
	area atRow: 6 atColumn: 1 put: 1.
	area atRow: 6 atColumn: 2 put: 0.
	area atRow: 6 atColumn: 3 put: 0.
	area atRow: 6 atColumn: 4 put: 0.
	area atRow: 6 atColumn: 5 put: 2.
	area atRow: 6 atColumn: 6 put: 4.
	area atRow: 6 atColumn: 7 put: 0.
	area atRow: 7 atColumn: 1 put: 1.
	area atRow: 7 atColumn: 2 put: 1.
	area atRow: 7 atColumn: 3 put: 0.
	area atRow: 7 atColumn: 4 put: 0.
	area atRow: 7 atColumn: 5 put: 0.
	area atRow: 7 atColumn: 6 put: 0.
	area atRow: 7 atColumn: 7 put: 0.
	 ^ instance! !

!Map methodsFor: 'accessors' stamp: 'MorganePigny 1/4/2017 10:33'!
area
	^ area ! !

!Map methodsFor: 'calculating' stamp: 'MorganePigny 1/16/2017 16:00'!
astar
	"algorithm looping the search of the least cost path leading the robot to the goal"

	"declaration & initialization of the variables"

	| aNodeStart aNode aPath neighbours aCounter |
	aNodeStart := NodeStart new withStartMap: self.
	aPath := Paths new.
	neighbours := Array new: 4.
	aNode := Node new withStartNode: aNodeStart andMap: self.
	aCounter := 0.

	"Adding NodeStart to the SortedColl: nodesToVisit"
	aPath nodesToVisit add: aNode.

	"While nodesToVisit is not empty and the robot is not on the goal"
	[ aPath nodesToVisit isNotEmpty & self solved not ]
		whileTrue: [ aNode := aPath nodesToVisit at: 1.	"we get the first element of nodesToVisit"

			"we remove it from nodesToVisit"
			aPath removeFromNodesToVisit: aNode.

			"we look if it's the final state"
			(self isRobotOnGoal: aNode)
				ifTrue: [ self solved: true ]
				ifFalse: [ (aNode isNewStateFrom: aPath)
						ifTrue: [ 
							neighbours := aNode getNeighboursIn: self withAPath: aPath.

							"we push these neighbours in the queue of nodesToVisit"
							aPath addNeighboursToTheSortedCollection: neighbours.

							"we put the new paths in the pathsTab; if it's the first loop we 'initialize' the pathsTab"
							aCounter == 0
								ifTrue: [ aPath addNeighboursForTheFirstTime: neighbours ofTheNodeToThePathsTab: aNode ]
								ifFalse: [ aPath addNeighbours: neighbours ofTheNodeToThePathsTab: aNode ].

							"we increment aCounter"
							aCounter := aCounter + 1.

							"we set the node as visited"
							aPath setVisited: aNode ]

					"we get the neighbours of the node and put the node as visited" ]
			"if so we get out of the loop"
			"if not we check if it's an already visited state or not" ].
	self solved
		ifTrue: [ 'Astar finished !! ' asMorph openInWindow.
			self getTheSolutionPath: aPath ]
		ifFalse: [ 'Could not find the solution...' asMorph openInWindow ]! !

!Map methodsFor: 'initialize' stamp: 'MorganePigny 1/16/2017 16:00'!
fillLevel1
	area atRow: 1 atColumn: 1 put: 1.
	area atRow: 1 atColumn: 2 put: 0.
	area atRow: 1 atColumn: 3 put: 0.
	area atRow: 1 atColumn: 4 put: 0.
	area atRow: 1 atColumn: 5 put: 0.
	area atRow: 1 atColumn: 6 put: 0.
	area atRow: 1 atColumn: 7 put: 1.
	area atRow: 2 atColumn: 1 put: 0.
	area atRow: 2 atColumn: 2 put: 3.
	area atRow: 2 atColumn: 3 put: 1.
	area atRow: 2 atColumn: 4 put: 0.
	area atRow: 2 atColumn: 5 put: 0.
	area atRow: 2 atColumn: 6 put: 0.
	area atRow: 2 atColumn: 7 put: 1.
	area atRow: 3 atColumn: 1 put: 0.
	area atRow: 3 atColumn: 2 put: 1.
	area atRow: 3 atColumn: 3 put: 1.
	area atRow: 3 atColumn: 4 put: 0.
	area atRow: 3 atColumn: 5 put: 0.
	area atRow: 3 atColumn: 6 put: 0.
	area atRow: 3 atColumn: 7 put: 0.
	area atRow: 4 atColumn: 1 put: 0.
	area atRow: 4 atColumn: 2 put: 2.
	area atRow: 4 atColumn: 3 put: 0.
	area atRow: 4 atColumn: 4 put: 0.
	area atRow: 4 atColumn: 5 put: 0.
	area atRow: 4 atColumn: 6 put: 0.
	area atRow: 4 atColumn: 7 put: 0.
	area atRow: 5 atColumn: 1 put: 0.
	area atRow: 5 atColumn: 2 put: 0.
	area atRow: 5 atColumn: 3 put: 0.
	area atRow: 5 atColumn: 4 put: 0.
	area atRow: 5 atColumn: 5 put: 0.
	area atRow: 5 atColumn: 6 put: 0.
	area atRow: 5 atColumn: 7 put: 0.
	area atRow: 6 atColumn: 1 put: 1.
	area atRow: 6 atColumn: 2 put: 0.
	area atRow: 6 atColumn: 3 put: 0.
	area atRow: 6 atColumn: 4 put: 0.
	area atRow: 6 atColumn: 5 put: 2.
	area atRow: 6 atColumn: 6 put: 4.
	area atRow: 6 atColumn: 7 put: 0.
	area atRow: 7 atColumn: 1 put: 1.
	area atRow: 7 atColumn: 2 put: 1.
	area atRow: 7 atColumn: 3 put: 0.
	area atRow: 7 atColumn: 4 put: 0.
	area atRow: 7 atColumn: 5 put: 0.
	area atRow: 7 atColumn: 6 put: 0.
	area atRow: 7 atColumn: 7 put: 0! !

!Map methodsFor: 'initialize' stamp: 'MorganePigny 1/11/2017 17:17'!
fillUp
	self fillLevel1.
	pos_robot := area indexOf: 3.
	pos_goal := area indexOf: 4.
	"we put a 0 instead of the 3 for the position of the robot in order to make calculation easier later"
	area replaceAll: 3 with: 0! !

!Map methodsFor: 'accessors' stamp: 'MorganePigny 1/16/2017 15:44'!
getTheSolutionPath: aPath
	aPath pathsTab
		do: [ :linkedList | 
			linkedList last pos_relative x == self pos_goal x
				ifTrue: [ linkedList last pos_relative x == self pos_goal x
						ifTrue: [ self solutionPath: linkedList ] ] ]! !

!Map methodsFor: 'initialize' stamp: 'MorganePigny 1/16/2017 15:44'!
initialize
	super initialize.
	area := CTGrid new: 7.
	solutionPath := LinkedList new.
	solved := false! !

!Map methodsFor: 'tests' stamp: 'MorganePigny 1/16/2017 14:52'!
isRobotOnGoal: aNode
	
	aNode pos_relative x == self pos_goal x ifTrue: [ aNode pos_relative y == self pos_goal y 
		ifTrue: [ ^ true ]
		ifFalse: [ false ] ].
	
	^ false! !

!Map methodsFor: 'accessors' stamp: 'MorganePigny 1/6/2017 16:21'!
pos_goal
	^ pos_goal! !

!Map methodsFor: 'accessors' stamp: 'MorganePigny 1/6/2017 16:21'!
pos_robot
	^ pos_robot ! !

!Map methodsFor: 'accessors' stamp: 'MorganePigny 1/16/2017 15:46'!
solutionPath

	^ solutionPath ! !

!Map methodsFor: 'accessors' stamp: 'MorganePigny 1/16/2017 15:43'!
solutionPath: aLinkedList
	solutionPath := aLinkedList postCopy.! !

!Map methodsFor: 'accessors' stamp: 'MorganePigny 1/10/2017 14:15'!
solved
	^ solved ! !

!Map methodsFor: 'accessors' stamp: 'MorganePigny 1/10/2017 14:16'!
solved: aBool
	
	solved := aBool.! !

Map subclass: #NodeStart
	instanceVariableNames: 'pos_relative totalCost depCost heuristicCost'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Astar'!

!NodeStart commentStamp: 'MorganePigny 1/9/2017 10:53' prior: 0!
I am the starting node of all the nodes, the one at the starting position of the Robot. Thus, I don't have a previous node.!

NodeStart subclass: #Node
	instanceVariableNames: 'previousNode'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Astar'!

!Node commentStamp: 'MorganePigny 1/4/2017 10:00' prior: 0!
I'm a node representing the current configuration of the world and the cost associated to get there.!

!Node methodsFor: 'comparing' stamp: 'MorganePigny 1/12/2017 13:54'!
== aNode

	self pos_relative == aNode pos_relative ifTrue: [ self previousNode == aNode previousNode ifTrue: [ ^ true ] ]
														ifFalse: [ ^ false ]! !

!Node methodsFor: 'calculating' stamp: 'MorganePigny 1/16/2017 14:32'!
calculatingDepCostFrom: aMap
	"the cost equals the cost of the previous node plus the cost of the robot moving from this previous node to the current one (1 if it goes through the ground, 5 if it's through a trap)"

	| newCost |
	newCost := self previousNode depCost.

	"current cost"
	(aMap area atRow: self pos_relative y atColumn: self pos_relative x) == 0
		ifTrue: [ newCost := newCost + 1 ].
	(aMap area atRow: self pos_relative y atColumn: self pos_relative x) == 2
		ifTrue: [ newCost := newCost + 5 ].

	"total cost"
	self depCost: newCost! !

!Node methodsFor: 'calculating' stamp: 'MorganePigny 1/16/2017 14:18'!
calculatingHeuristicFrom: aMap
	"we are using the Manhattan heuristic which is the distance from the robot to the goal"

	| row_diff col_diff |
	row_diff := (self pos_relative y - aMap pos_goal y) abs.
	col_diff := (self pos_relative x - aMap pos_goal x) abs.
	
	self heuristicCost: (row_diff + col_diff).! !

!Node methodsFor: 'calculating' stamp: 'MorganePigny 1/16/2017 14:34'!
calculatingTotalCostFrom: aMap
	self calculatingHeuristicFrom: aMap.
	self calculatingDepCostFrom: aMap.
	self totalCost: self depCost + self heuristicCost! !

!Node methodsFor: 'moving' stamp: 'MorganePigny 1/12/2017 15:09'!
checkingDown: aMap
	"check if we can go down (if there is no wall)"
	
	| ar |
	ar := aMap area.
	
	(self pos_relative y + 1) <= (ar rowCount) ifTrue: [ (ar atRow: self pos_relative y + 1 atColumn: self pos_relative x) ~= 1 
		ifTrue: [ ^ true ]	
	].

	^ false

	
	 ! !

!Node methodsFor: 'moving' stamp: 'MorganePigny 1/16/2017 13:48'!
checkingDown: aMap withAPath: aPath
	"check if we can go down (if there is no wall)"

	| ar n alreadyVisited |
	ar := aMap area.
	n := Node new.
	n := n newNodeCopy: self.
	alreadyVisited := false.
	n setX: self pos_relative x setY: self pos_relative y + 1.
	self pos_relative y + 1 <= ar rowCount
		ifTrue: [ (ar atRow: self pos_relative y + 1 atColumn: self pos_relative x) ~= 1
				ifTrue: [ aPath nodesVisited isEmptyOrNil
						ifTrue: [ ^ true ]
						ifFalse: [ 
							aPath nodesVisited
								do: [ :each | 
									each pos_relative x == n pos_relative x
										ifTrue: [ each pos_relative y == n pos_relative y
												ifTrue: [ ^ false ] ] ].
							^ true ] ] ].
	^ false! !

!Node methodsFor: 'moving' stamp: 'MorganePigny 1/12/2017 15:09'!
checkingLeft: aMap
	"check if the robot can go to the left"
	
		| ar |
	ar := aMap area.
	
		(self pos_relative x - 1) > 0 ifTrue: [ (ar atRow: self pos_relative y atColumn: self pos_relative x - 1) ~= 1 ifTrue: [ ^ true ] ].
		
		^ false
	
! !

!Node methodsFor: 'moving' stamp: 'MorganePigny 1/16/2017 14:04'!
checkingLeft: aMap withAPath: aPath
	"check if the robot can go to the left"

	| ar n alreadyVisited |
	ar := aMap area.
	n := Node new.
	n := n newNodeCopy: self.
	alreadyVisited := false.
	n setX: self pos_relative x - 1 setY: self pos_relative y.
	self pos_relative x - 1 > 0
		ifTrue: [ (ar atRow: self pos_relative y atColumn: self pos_relative x - 1) ~= 1
				ifTrue: [ aPath nodesVisited isEmptyOrNil
						ifTrue: [ ^ true ]
						ifFalse: [ aPath nodesVisited
								do: [ :each | 
									each pos_relative x == n pos_relative x
										ifTrue: [ each pos_relative y == n pos_relative y
												ifTrue: [ ^ false ] ] ].
							^ true ] ] ].
	^ false! !

!Node methodsFor: 'moving' stamp: 'MorganePigny 1/12/2017 15:09'!
checkingRight: aMap
	"check if the robot can go to the right"
	
		| ar |
	ar := aMap area.
	
	(self pos_relative x + 1) <= (ar columnCount) ifTrue: [ (ar atRow: self pos_relative y atColumn: self pos_relative x + 1) ~= 1 ifTrue: [ ^ true ] ].
	
	^ false
! !

!Node methodsFor: 'moving' stamp: 'MorganePigny 1/16/2017 14:04'!
checkingRight: aMap withAPath: aPath
	"check if the robot can go to the right"

	| ar n alreadyVisited |
	ar := aMap area.
	n := Node new.
	n := n newNodeCopy: self.
	alreadyVisited := false.
	n setX: self pos_relative x + 1 setY: self pos_relative y.
	self pos_relative x + 1 <= ar columnCount
		ifTrue: [ (ar atRow: self pos_relative y atColumn: self pos_relative x + 1) ~= 1
				ifTrue: [ aPath nodesVisited isEmptyOrNil
						ifTrue: [ ^ true ]
						ifFalse: [ aPath nodesVisited
								do: [ :each | 
									each pos_relative x == n pos_relative x
										ifTrue: [ each pos_relative y == n pos_relative y
												ifTrue: [ ^ false ] ] ].
							^ true ] ] ].
	^ false! !

!Node methodsFor: 'moving' stamp: 'MorganePigny 1/12/2017 15:09'!
checkingUp: aMap
	"check if the robot can go up"
	
		| ar |
	ar := aMap area.
	
		(self pos_relative y - 1) > 0 ifTrue: [ (ar atRow: self pos_relative y - 1 atColumn: self pos_relative x) ~= 1 ifTrue: [ ^ true ] ].
		
		^ false
			! !

!Node methodsFor: 'moving' stamp: 'MorganePigny 1/16/2017 14:04'!
checkingUp: aMap withAPath: aPath
	"check if the robot can go up"

	| ar n alreadyVisited |
	ar := aMap area.
	n := Node new.
	n := n newNodeCopy: self.
	alreadyVisited := false.
	n setX: self pos_relative x setY: self pos_relative y - 1.
	self pos_relative y - 1 > 0
		ifTrue: [ (ar atRow: self pos_relative y - 1 atColumn: self pos_relative x) ~= 1
				ifTrue: [ aPath nodesVisited isEmptyOrNil
						ifTrue: [ ^ true ]
						ifFalse: [ aPath nodesVisited
								do: [ :each | 
									each pos_relative x == n pos_relative x
										ifTrue: [ each pos_relative y == n pos_relative y
												ifTrue: [ ^ false ] ] ].
							^ true ] ] ].
	^ false! !

!Node methodsFor: 'initialize' stamp: 'MorganePigny 1/11/2017 17:26'!
initialize
	| prev |
	prev := NodeStart new.
	previousNode := prev.
	totalCost := 0.
	depCost := 0.
	heuristicCost := 0.
	pos_relative := pos_robot.
	nodesVisited := OrderedCollection new.
	nodesVisited add: prev! !

!Node methodsFor: 'comparing' stamp: 'MorganePigny 1/13/2017 14:06'!
isEmptyOrNil

	^ false! !

!Node methodsFor: 'initialize' stamp: 'MorganePigny 1/13/2017 13:42'!
newNodeCopy: aNode

	| newN |
	
	super copy.
	
	newN := Node new.
	
	newN pos_relative: aNode pos_relative copy.
	newN totalCost: aNode totalCost copy.
	newN depCost: aNode depCost copy.
	newN heuristicCost: aNode heuristicCost copy.
	newN previousNode: aNode previousNode copy.
	
	^ newN
	
	! !

!Node methodsFor: 'accessors' stamp: 'MorganePigny 1/12/2017 15:30'!
pos_goal
	
	^ super pos_goal! !

!Node methodsFor: 'accessors' stamp: 'MorganePigny 1/9/2017 11:03'!
previousNode
	^ previousNode ! !

!Node methodsFor: 'accessors' stamp: 'MorganePigny 1/13/2017 09:29'!
previousNode: aNode
	
	previousNode := aNode ! !

!Node methodsFor: 'initialize' stamp: 'MorganePigny 1/12/2017 15:44'!
withStartNode: aNodeStart 
	super withStartMap: aNodeStart.	"pb? envoyer une map à la place?"
	previousNode := aNodeStart.
	nodesVisited add: aNodeStart! !

!Node methodsFor: 'initialize' stamp: 'MorganePigny 1/12/2017 15:43'!
withStartNode: aNodeStart andMap: aMap
	super withStartMap: aMap.	"pb? envoyer une map à la place?"
	previousNode := aNodeStart.
	nodesVisited add: aNodeStart! !

!NodeStart methodsFor: 'neighbours' stamp: 'MorganePigny 1/12/2017 10:13'!
addNeighbours: anArray toTheSortedCollectionOfThePath: aPath
	"put the array of neighbours into the SortedCollection"

	 anArray do: [ :each | each notNil
				ifTrue: [ aPath addNodeToTheSortedCollection: each ] ].
	^ aPath! !

!NodeStart methodsFor: 'neighbours' stamp: 'MorganePigny 1/12/2017 09:55'!
calculateNeighboursCost: anArray
	"calculate the cost of each neighbour from the current node"

	anArray
		do: [ :each | 
			each notNil
				ifTrue: [ each calculatingTotalCost ] ].
	^ anArray! !

!NodeStart methodsFor: 'accessors' stamp: 'MorganePigny 1/12/2017 15:41'!
depCost
	^ depCost! !

!NodeStart methodsFor: 'accessors' stamp: 'MorganePigny 1/13/2017 09:30'!
depCost: aCost

	depCost := aCost! !

!NodeStart methodsFor: 'neighbours' stamp: 'MorganePigny 1/16/2017 14:34'!
getNeighbourDownIn: aMap
	"return an array of Nodes corresponding to the new nodes created from to the current one"
	
	"if we manage to create a new node from the current one (which means that the node we get is different from our current one"
	(self checkingDown: aMap) 
		ifTrue: [ self pos_relative setX: self pos_relative x setY: self pos_relative y + 1 ].

	! !

!NodeStart methodsFor: 'neighbours' stamp: 'MorganePigny 1/16/2017 14:34'!
getNeighbourDownIn: aMap withAPath: aPath
	"return an array of Nodes corresponding to the new nodes created from to the current one"
	
	"if we manage to create a new node from the current one (which means that the node we get is different from our current one"
	(self checkingDown: aMap withAPath: aPath) 
		ifTrue: [ self pos_relative setX: self pos_relative x setY: self pos_relative y + 1].

	! !

!NodeStart methodsFor: 'neighbours' stamp: 'MorganePigny 1/16/2017 14:34'!
getNeighbourLeftIn: aMap
	"return an array of Nodes corresponding to the new nodes created from to the current one"

	"if we manage to create a new node from the current one (which means that the node we get is different from our current one"

	(self checkingLeft: aMap)
		ifTrue: [ self pos_relative setX: self pos_relative x - 1 setY: self pos_relative y ]! !

!NodeStart methodsFor: 'neighbours' stamp: 'MorganePigny 1/16/2017 14:34'!
getNeighbourLeftIn: aMap withAPath: aPath
	"return an array of Nodes corresponding to the new nodes created from to the current one"

	"if we manage to create a new node from the current one (which means that the node we get is different from our current one"

	(self checkingLeft: aMap withAPath: aPath)
		ifTrue: [ self pos_relative setX: self pos_relative x - 1 setY: self pos_relative y ]! !

!NodeStart methodsFor: 'neighbours' stamp: 'MorganePigny 1/16/2017 14:34'!
getNeighbourRightIn: aMap
	"return an array of Nodes corresponding to the new nodes created from to the current one"

	"if we manage to create a new node from the current one (which means that the node we get is different from our current one"
	(self checkingRight: aMap)
		ifTrue: [ self pos_relative setX: self pos_relative x + 1 setY: self pos_relative y].

	! !

!NodeStart methodsFor: 'neighbours' stamp: 'MorganePigny 1/16/2017 14:34'!
getNeighbourRightIn: aMap withAPath: aPath
	"return an array of Nodes corresponding to the new nodes created from to the current one"

	"if we manage to create a new node from the current one (which means that the node we get is different from our current one"

	(self checkingRight: aMap withAPath: aPath)
		ifTrue: [ self pos_relative setX: self pos_relative x + 1 setY: self pos_relative y]! !

!NodeStart methodsFor: 'neighbours' stamp: 'MorganePigny 1/16/2017 14:35'!
getNeighbourUpIn: aMap
	"return an array of Nodes corresponding to the new nodes created from to the current one"

	"if we manage to create a new node from the current one (which means that the node we get is different from our current one"

	(self checkingUp: aMap)
		ifTrue: [ self pos_relative setX: self pos_relative x setY: self pos_relative y - 1 ]! !

!NodeStart methodsFor: 'neighbours' stamp: 'MorganePigny 1/16/2017 14:35'!
getNeighbourUpIn: aMap withAPath: aPath
	"return an array of Nodes corresponding to the new nodes created from to the current one"

	"if we manage to create a new node from the current one (which means that the node we get is different from our current one"

	(self checkingUp: aMap withAPath: aPath)
		ifTrue: [ self pos_relative setX: self pos_relative x setY: self pos_relative y - 1 ]! !

!NodeStart methodsFor: 'neighbours' stamp: 'MorganePigny 1/13/2017 15:35'!
getNeighboursIn: aMap
	"return an array of Nodes corresponding to the new nodes created from to the current one"

	| anArray n1 n2 n3 n4 |
	anArray := Array new: 4.
	n1 := Node new.
	
	n1 := n1 newNodeCopy: self.
	n2 := Node new.
	n2 := n2 newNodeCopy: self.
	n3 := Node new.
	n3 := n3 newNodeCopy: self.
	n4 := Node new.
	n4 := n4 newNodeCopy: self.

	"we get a potential new node above the current one"
	(n1 getNeighbourUpIn: aMap) pos_relative y == self pos_relative y ifFalse: [ anArray at: 1 put: n1 ].
	
	"we do the same thing below"
	(n2 getNeighbourDownIn: aMap) pos_relative y == self pos_relative y ifFalse: [ anArray at: 2 put: n2 ].
	
	"to the left"
	(n3 getNeighbourLeftIn: aMap) pos_relative x == self pos_relative x ifFalse: [ anArray at: 3 put: n3 ].
	
	"to the right"
	(n4 getNeighbourDownIn: aMap) pos_relative x == self pos_relative x ifFalse: [ anArray at: 4 put: n4 ].
	
	^ anArray! !

!NodeStart methodsFor: 'neighbours' stamp: 'MorganePigny 1/16/2017 15:55'!
getNeighboursIn: aMap withAPath: aPath
	"return an array of Nodes corresponding to the new nodes created from to the current one"

	| anArray n1 n2 n3 n4 |
	anArray := Array new: 4.
	n1 := Node new.
	n1 := n1 newNodeCopy: self.
	n2 := Node new.
	n2 := n2 newNodeCopy: self.
	n3 := Node new.
	n3 := n3 newNodeCopy: self.
	n4 := Node new.
	n4 := n4 newNodeCopy: self.

	"we get a potential new node above the current one"
	(n1 getNeighbourUpIn: aMap withAPath: aPath) pos_relative y == self pos_relative y
		ifFalse: [ n1 previousNode: self.
			n1 calculatingTotalCostFrom: aMap.
			anArray at: 1 put: n1 ].
	
	
	"we do the same thing below"
	(n2 getNeighbourDownIn: aMap withAPath: aPath) pos_relative y == self pos_relative y
		ifFalse: [ n2 previousNode: self.
			n2 calculatingTotalCostFrom: aMap.
			anArray at: 2 put: n2 ].

	"to the left"
	(n3 getNeighbourLeftIn: aMap withAPath: aPath) pos_relative x == self pos_relative x
		ifFalse: [ n3 previousNode: self.
			n3 calculatingTotalCostFrom: aMap.
			anArray at: 3 put: n3 ].

	"to the right"
	(n4 getNeighbourRightIn: aMap withAPath: aPath) pos_relative x == self pos_relative x
		ifFalse: [ n4 previousNode: self.
			n4 calculatingTotalCostFrom: aMap.
			anArray at: 4 put: n4 ].
	^ anArray! !

!NodeStart methodsFor: 'accessors' stamp: 'MorganePigny 1/12/2017 15:41'!
heuristicCost
	^ heuristicCost! !

!NodeStart methodsFor: 'accessors' stamp: 'MorganePigny 1/13/2017 09:31'!
heuristicCost: aCost

	heuristicCost := aCost! !

!NodeStart methodsFor: 'initalize' stamp: 'MorganePigny 1/12/2017 15:43'!
initialize
	
	super initialize.
	pos_relative := pos_robot.
	nodesVisited := OrderedCollection new.
	depCost := 0.
	heuristicCost := 0.
	totalCost := 0.
! !

!NodeStart methodsFor: 'testing' stamp: 'MorganePigny 1/16/2017 14:59'!
isNewStateFrom: aPath
	"we check if this node was already visited"

	aPath nodesVisited isEmptyOrNil
		ifTrue: [ ^ true ]
		ifFalse: [ aPath nodesVisited
				do: [ :each | 
					each pos_relative x == self pos_relative x
						ifTrue: [ each pos_relative y == self pos_relative y
								ifTrue: [ ^ false ] ] ].
			^ true ].
	! !

!NodeStart methodsFor: 'initalize' stamp: 'MorganePigny 1/13/2017 11:36'!
newNodeCopy: aNodeS
	
	| newNS |
	
	super copy.
	
	newNS := NodeStart new.
	
	newNS pos_relative: aNodeS pos_relative copy.
	newNS totalCost: aNodeS totalCost copy.
	newNS depCost: aNodeS depCost copy.
	newNS heuristicCost: aNodeS heuristicCost copy.
	
	^ newNS
	
	! !

!NodeStart methodsFor: 'accessors' stamp: 'MorganePigny 1/12/2017 15:30'!
pos_goal
	
	^ super pos_goal! !

!NodeStart methodsFor: 'accessors' stamp: 'MorganePigny 1/9/2017 11:04'!
pos_relative
	^ pos_relative! !

!NodeStart methodsFor: 'accessors' stamp: 'MorganePigny 1/10/2017 14:26'!
pos_relative: aPos
	
	pos_relative := aPos! !

!NodeStart methodsFor: 'printing' stamp: 'MorganePigny 1/11/2017 16:59'!
printOn: astream
	super printOn: astream.
	astream nextPutAll: ' ', pos_relative asString! !

!NodeStart methodsFor: 'accessors' stamp: 'MorganePigny 1/13/2017 09:46'!
setX: aPosx setY: aPosy
	
	self pos_relative setX: aPosx setY: aPosy.! !

!NodeStart methodsFor: 'accessors' stamp: 'MorganePigny 1/10/2017 10:22'!
totalCost
	^ totalCost! !

!NodeStart methodsFor: 'accessors' stamp: 'MorganePigny 1/13/2017 09:30'!
totalCost: aCost

	totalCost := aCost! !

!NodeStart methodsFor: 'initalize' stamp: 'MorganePigny 1/12/2017 15:43'!
withStartMap: aMap
	pos_relative := aMap pos_robot.
	nodesVisited := OrderedCollection new.
	depCost := 0.
	heuristicCost := 0.
	totalCost := 0! !

Map subclass: #Paths
	instanceVariableNames: 'nodesToVisit pathsTab nodesVisited'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Astar'!

!Paths commentStamp: 'MorganePigny 1/4/2017 09:58' prior: 0!
I'm a robot with an assigned position and goal that I want to reach by going through the least cost path.!

!Paths methodsFor: 'adding/removing' stamp: 'MorganePigny 1/16/2017 13:54'!
addNeighbours: anArray ofTheNodeToThePathsTab: aNode
	"add the new paths leading to my neighbours found previously"

	| aLinkedList aTab aLL2 aLL3 aLL4 |
	aLinkedList := LinkedList new.
	aLL2 := LinkedList new.
	aLL3 := LinkedList new.
	aLL4 := LinkedList new.
	aTab := OrderedCollection new.

	"for each already existing linked list leading to my node (which means that their last element is my node), we create *size of neighbours* new linked lists leading to the neighbours of my node found previously"
	pathsTab
		do: [ :each | 
			each last == aNode
				ifTrue: [ (anArray at: 1) isEmptyOrNil
						ifFalse: [ aLinkedList := each copyWith: (anArray at: 1).
							aTab add: aLinkedList ].
					(anArray at: 2) isEmptyOrNil
						ifFalse: [ aLL2 := each copyWith: (anArray at: 2).
							aTab add: aLL2 ].
					(anArray at: 3) isEmptyOrNil
						ifFalse: [ aLL3 := each copyWith: (anArray at: 3).
							aTab add: aLL3 ].
					(anArray at: 4) isEmptyOrNil
						ifFalse: [ aLL4 := each copyWith: (anArray at: 4).
							aTab add: aLL4 ] ] ].
				
	aTab do: [ :each | pathsTab add: each ]! !

!Paths methodsFor: 'adding/removing' stamp: 'MorganePigny 1/13/2017 14:26'!
addNeighboursForTheFirstTime: anArray ofTheNodeToThePathsTab: aNode 
	"add the new paths leading to my neighbours found previously"
	
	| aLinkedList aLL2 aLL3 aLL4 |
	aLinkedList := LinkedList new.
	aLL2 := LinkedList new.
	aLL3 := LinkedList new.
	aLL4 := LinkedList new.
	
	"we create linked list for the neighbours found with the Node star as first element, and the neighbour found as the end of the linked list"
	(anArray at: 1) isEmptyOrNil ifFalse: [ 
			aLinkedList add: aNode. aLinkedList add: (anArray at: 1) after: aNode. 
			self pathsTab add: aLinkedList ].
		
	(anArray at: 2) isEmptyOrNil ifFalse: [ 
			aLL2 add: aNode. aLL2 add: (anArray at: 2) after: aNode. 
			self pathsTab add: aLL2 ].
		
	(anArray at: 3) isEmptyOrNil ifFalse: [ 
			aLL3 add: aNode. aLL3 add: (anArray at: 3) after: aNode. 
			self pathsTab add: aLL3 ].
		
	(anArray at: 4) isEmptyOrNil ifFalse: [ 
			aLL4 add: aNode. aLL4 add: (anArray at: 4) after: aNode. 
			self pathsTab add: aLL4 ].
	
! !

!Paths methodsFor: 'adding/removing' stamp: 'MorganePigny 1/12/2017 10:18'!
addNeighboursToTheSortedCollection: anArray
	"add an array of neighbours to the sorted collection"
	
	 anArray do: [ :each | each notNil
				ifTrue: [ self nodesToVisit add: each ] ].
 	! !

!Paths methodsFor: 'adding/removing' stamp: 'MorganePigny 1/10/2017 16:10'!
addSonNode: aSNode ToTheNode: aNode
	"add a path to the array of linked lists by linking a son node to its father"
	
	"| aLinkedList |
	
	aLinkedList := LinkedList new.
	aLinkedList add: aSNode after: aNode."
	
	"we need to see if there is an other linked list implemented in the array which has for last element our first element => which means it's a path"
	"pathsTab do: [ :each | each last == aNode ifTrue: [ each add: aNode afterLink: each ] ]"! !

!Paths methodsFor: 'initialize' stamp: 'MorganePigny 1/16/2017 15:44'!
initialize
	super initialize.
	pathsTab := OrderedCollection new.
	nodesToVisit := SortedCollection sortBlock: [ :n1 :n2 | n1 totalCost <= n2 totalCost ].
	nodesVisited := OrderedCollection new.
! !

!Paths methodsFor: 'accessors' stamp: 'MorganePigny 1/10/2017 09:55'!
nodesToVisit
	^ nodesToVisit ! !

!Paths methodsFor: 'accessors' stamp: 'MorganePigny 1/10/2017 11:27'!
nodesToVisit: aSortedCollection
	
	nodesToVisit := aSortedCollection ! !

!Paths methodsFor: 'accessors' stamp: 'MorganePigny 1/16/2017 10:20'!
nodesVisited
	^ nodesVisited! !

!Paths methodsFor: 'accessors' stamp: 'MorganePigny 1/10/2017 09:55'!
pathsTab
	^ pathsTab ! !

!Paths methodsFor: 'accessors' stamp: 'MorganePigny 1/10/2017 11:27'!
pathsTab: anArrayOfLinkedLists
	
	pathsTab := anArrayOfLinkedLists ! !

!Paths methodsFor: 'adding/removing' stamp: 'MorganePigny 1/12/2017 10:52'!
removeFromNodesToVisit: aNode

	nodesToVisit remove: aNode ifAbsent: [  ].! !

!Paths methodsFor: 'accessors' stamp: 'MorganePigny 1/16/2017 09:35'!
setVisited: aNode
	
	nodesVisited add: aNode.! !

TestCase subclass: #NodeTest
	instanceVariableNames: 'pos_relative totalCost previousNode'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Astar'!

!NodeTest commentStamp: 'MorganePigny 1/6/2017 16:49' prior: 0!
I am a class testing the Node class!

!NodeTest methodsFor: 'tests' stamp: 'MorganePigny 1/9/2017 15:01'!
setUp
	pos_relative := 2@2.
	totalCost := 0

! !

!NodeTest methodsFor: 'tests' stamp: 'MorganePigny 1/12/2017 14:55'!
testGoingDown
	
	| ar pos n ns |
	ar := Map new fillUp.
	ns := NodeStart new withStartMap: ar.
	n := Node new withStartNode: ns andMap: ar.
	pos := n pos_relative y + 1.
	
	self assert: (n checkingDown: ar) pos_relative y equals: pos.
! !

!NodeTest methodsFor: 'tests' stamp: 'MorganePigny 1/12/2017 15:00'!
testGoingLeft
	
	| ar pos n ns |
	ar := Map new fillUp.
	ns := NodeStart new withStartMap: ar.
	n := Node new withStartNode: ns andMap: ar.
	pos := n pos_relative x - 1.
	
	self assert: (n checkingLeft: ar) pos_relative x equals: pos.
! !

!NodeTest methodsFor: 'tests' stamp: 'MorganePigny 1/12/2017 15:01'!
testGoingRight
	
	| ar pos n ns |
	ar := Map new fillUp.
	ns := NodeStart new withStartMap: ar.
	n := Node new withStartNode: ns andMap: ar.
	pos := n pos_relative x + 1.
	
	self assert: (n checkingRight: ar) pos_relative x equals: pos.
! !

!NodeTest methodsFor: 'tests' stamp: 'MorganePigny 1/12/2017 14:57'!
testGoingUp
	
	| ar pos n ns |
	ar := Map new fillUp.
	ns := NodeStart new withStartMap: ar.
	n := Node new withStartNode: ns andMap: ar.
	pos := n pos_relative y - 1.
	
	self assert: (n checkingUp: ar) pos_relative y equals: pos.
! !

!NodeTest methodsFor: 'tests' stamp: 'MorganePigny 1/12/2017 15:44'!
testNeighboursDifferentFromCurrentNode
	| ar ns n |
	ar := Map new fillUp.
	ns := NodeStart new withStartMap: ar.
	n := Node new withStartNode: ns.
	self assert: ((ns getNeighboursIn: ar) identityIndexOf: n) equals: 0! !

!NodeTest methodsFor: 'tests' stamp: 'MorganePigny 1/13/2017 09:40'!
testNewNode
	
	| n1 n2 n3 ns|
	ns := NodeStart new.
	n1 := Node new withStartNode: ns.
	n2 := Node new withStartNode: ns.
	n3 := Node new withStartNode: ns.
	
	n2 totalCost: 5.
	n2 heuristicCost: 3.
	n2 depCost: 2.
	n2 pos_relative: 4@4.
	n2 previousNode: n3.
	
	n1 newNode: n2.
	
	self assert: n1 totalCost equals: 5.
	self assert: n1 heuristicCost equals: 3.
	self assert: n1 depCost equals: 2.
	self assert: n1 pos_relative equals: 4@4.
	self assert: n1 previousNode equals: n3.
	
! !

!NodeTest methodsFor: 'tests' stamp: 'MorganePigny 1/9/2017 13:18'!
testNodeOnMap
	
	| ar |
	ar := Map map1.
	self assert: (pos_relative x between: 0 and: ar area columnCount).
	self assert: (pos_relative y between: 0 and: ar area rowCount)! !

!NodeTest methodsFor: 'tests' stamp: 'MorganePigny 1/11/2017 17:12'!
testNodeTotalCostSuperiorToPreviousNode
	| n |
	n := Node new.
	self assert: n totalCost >= n previousNode totalCost! !

!NodeTest methodsFor: 'tests' stamp: 'MorganePigny 1/11/2017 17:12'!
testTotalCost
	| n |
	n := Node new.
	self assert: n totalCost equals: n depCost + n heuristicCost! !

TestCase subclass: #PathsTest
	instanceVariableNames: 'nodesToVisit pathsTab nodesVisited'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Astar'!

!PathsTest commentStamp: 'MorganePigny 1/10/2017 09:56' prior: 0!
I am a class testing the class Paths!

!PathsTest methodsFor: 'tests' stamp: 'MorganePigny 1/16/2017 10:23'!
setUp
	| ns n m |
	m := Map new initialize.
	ns := NodeStart new withStartMap: m.
	n := Node new withStartNode: ns.
	nodesToVisit := SortedCollection sortBlock: [ :n1 :n2 | n1 totalCost <= n2 totalCost ].
	nodesToVisit addAll: { ns. n }.
	nodesVisited := OrderedCollection new.! !

!PathsTest methodsFor: 'tests' stamp: 'MorganePigny 1/16/2017 10:23'!
testNodesToVisitNotAlreadyVisited
	"we check if each nodes in the array is indeed not visited"


	nodesToVisit do: [ :each | self deny: (nodesVisited includes: each) ]! !

TestCase subclass: #RobotTest
	instanceVariableNames: 'pos_robot pos_goal'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Astar'!

!RobotTest commentStamp: 'MorganePigny 1/4/2017 10:28' prior: 0!
I'm a class used to make tests for the Robot!

!RobotTest methodsFor: 'tests' stamp: 'MorganePigny 1/4/2017 13:54'!
setUp
	pos_robot := 2@2.
	pos_goal := 6@6! !

!RobotTest methodsFor: 'tests' stamp: 'MorganePigny 1/4/2017 16:33'!
testRobotInMap
	
	| ar |
	ar := Map map1.
	self assert: (pos_robot x between: 0 and: ar area columnCount).
	self assert: (pos_robot y between: 0 and: ar area rowCount)! !

!RobotTest methodsFor: 'tests' stamp: 'MorganePigny 1/11/2017 17:17'!
testRobotOnGround
	| ar |
	ar := Map map1.
	ar fillUp.
	self
		assert: (ar area atRow: pos_robot y atColumn: pos_robot x)
		equals: 0! !

Morph subclass: #Environment
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Astar'!

!Environment commentStamp: 'MorganePigny 1/16/2017 16:23' prior: 0!
I'm the graphic representation of the Astar with one map, a robot and a goal.!

!Environment methodsFor: 'printing' stamp: 'MorganePigny 1/16/2017 16:34'!
drawOn: aCanvas
	| squareSize horizontalBar verticalBar |
	squareSize := self width.
	horizontalBar := self bounds insetBy: 0 @ squareSize.
	verticalBar := self bounds insetBy: squareSize @ 0.
	aCanvas fillRectangle: horizontalBar color: self color.
	aCanvas fillRectangle: verticalBar color: self color! !
